from cognite import expr
import mxnet as mx

def forward(activations, biases):
    output = mx.ndarray.add(activations, biases)
    def backward(gradients):
        return (gradients, gradients)
    return output, backward

def add(x, biases):
    if isinstance(x, expr.Constant) and isinstance(biases, expr.Constant):
        return expr.Constant(forward(x.value, biases.value)[0])
    else:
        return expr.Apply(expr.Function('add', forward), [x, biases])



import meta
import expr
import numpy as np

@meta.differentiable_function
def linear_layer(x, weights, biases):
    x = meta.linear(x, weights)
    x = meta.add(x, biases)
    return meta.relu(x)

output, backward = linear_layer(np.array([[1], [2]]), np.array([[2]]), np.array([3]))
print(output[0])
print(backward(np.array([[1], [1]])))



from cognite import expr
import numpy as np

def forward(a, b):
    difference = a - b
    output = difference**2
    def backwards(gradients):
        return (2*difference, -2*difference)
    return output, backwards

def squared_difference(x, weights):
    if isinstance(x, expr.Constant) and isinstance(weights, expr.Constant):
        return expr.Constant(forward(x.value, weights.value)[0])
    else:
        return expr.Apply(expr.Function('squared_difference', forward), [x, weights])



from cognite import expr
import mxnet as mx

def forward(x):
    mask = x > 0
    output = mx.ndarray.multiply(mask, x)
    def backward(gradient):
        return (mx.ndarray.multiply(gradient, mask),)
    return output, backward

def relu(x):
    if isinstance(x, expr.Constant):
        return expr.Constant(forward(x.value)[0])
    else:
        return expr.Apply(expr.Function('relu', forward), [x])



import collections

class Function:
    def __init__(self, name, fn):
        self.name = name
        self.fn = fn

    def __call__(self, *args):
        return self.fn(*args)

    def __repr__(self):
        return "Function(%s, %s)" % (repr(self.name), repr(self.fn))

class Expr:
    @property
    def children(self):
        return []

class Apply(Expr):
    def __init__(self, function, args):
        self.function = function
        self.args = args

    @property
    def children(self):
        return self.args

    def replace(self, replacements):
        new_args = [arg.replace(replacements) for arg in self.args]
        if new_args == self.args:
            return self
        elif all([isinstance(arg, Constant) for arg in new_args]):
            return Constant(self.function(*[arg.value for arg in new_args])[0])
        else:
            return Apply(self.function, new_args)

    def __repr__(self):
        return "Apply(%s, %s)" % (repr(self.function), repr(self.args))

class Constant(Expr):
    def __init__(self, value):
        self.value = value

    def replace(self, replacements):
        return self

    def __repr__(self):
        return repr(self.value)

class Variable(Expr):
    def __init__(self, name):
        self.name = name

    def replace(self, replacements):
        return replacements.get(self, self)

    def __repr__(self):
        return self.name

def topological_sort(root):
    exprs = []
    stack = [root]
    while stack:
        x = stack.pop()
        if not x in exprs:
            stack.extend(x.children)
            exprs.append(x)
    exprs.reverse()
    return exprs

def count_references(exprs):
    refs = collections.Counter()
    for expr in exprs:
        for c in expr.children:
            refs[c] += 1
    return refs



import argparse
import importlib

def load_value(name):
    parts = name.split('.')
    for i in reversed(range(1, len(parts) + 1)):
        try:
            module = importlib.import_module('.'.join(parts[:i]))
        except:
            pass
        else:
            output = module
            for x in range(i, len(parts)):
                output = getattr(output, parts[x])
            return output
    else:
        raise Exception('failed to find ' + name)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description = 'Training')

    parser.add_argument(
        '--data-loader',
        '-d',
        type=str,
        help='A expression for the data loader',
        required=True,
    )

    parser.add_argument(
        '--model',
        '-m',
        type=str,
        help='A expression for the model',
        required=True,
    )

    args = parser.parse_args()
    data_loader = load_value(args.data_loader)
    model = load_value(args.model)
    print(data_loader)
    print(model)



import os
from setuptools import setup

setup(
    name = "cognite",
    version = "0.0.1",
    author = "Daniel Waterworth",
    author_email = "daniel@manganizeme.com",
    description = "",
    keywords = "deep-learning",
    url = "https://github.com/danielwaterworth/cognite",
    packages=['cognite'],
    long_description="",
    classifiers=[],
    install_requires=['mxnet', 'numpy'],
)



import inspect

from cognite.linear import linear
from cognite.add import add
from cognite.relu import relu
from cognite.squared_difference import squared_difference
from cognite import combinators
from cognite import expr

class Function:
    def __init__(self, parameters, body):
        self.parameters = parameters
        self.body = body

    def __call__(self, *args):
        def f(x):
            if isinstance(x, expr.Expr):
                return x
            else:
                return expr.Constant(x)
        args = map(f, args)
        return self.body.replace(dict(zip(self.parameters, args)))

    def __repr__(self):
        return "Function(%s, %s)" % (repr(self.parameters), repr(self.body))

    def transform(self):
        exprs = expr.topological_sort(self.body)
        refs = expr.count_references(exprs)

        scope = list(self.parameters)
        operations = []

        mask = [refs[scope_e] == 0 for scope_e in scope]
        if True in mask:
            operations.append(
                combinators.Discard(
                    *[refs[scope_e] == 0 for scope_e in scope]
                )
            )
            scope = [scope_e for scope_e in scope if refs[scope_e] != 0]
            print(scope)

        for e in exprs:
            if not isinstance(e, expr.Variable):
                for child in e.children:
                    refs[child] -= 1

                if isinstance(e, expr.Apply):
                    mask = []
                    for scope_e in scope:
                        # If the variable is going to be used again, duplicate
                        # it
                        mask.append(
                            scope_e in e.args and refs[scope_e] != 0
                        )

                    if True in mask:
                        operations.append(combinators.Duplicate(*mask))
                        new_scope = []
                        for scope_e, dup in zip(scope, mask):
                            new_scope.append(scope_e)
                            if dup:
                                new_scope.append(scope_e)
                        scope = new_scope

                    indices = []
                    new_scope = []
                    for arg in e.args:
                        indices.append(scope.index(arg))
                        new_scope.append(arg)

                    for i, scope_e in enumerate(scope):
                        if not scope_e in e.args:
                            indices.append(i)
                            new_scope.append(scope_e)

                    if sorted(indices) != indices:
                        scope = new_scope
                        operations.append(
                            combinators.Permutation(
                                *indices
                            )
                        )

                    n = len(scope) - len(e.args)
                    if n == 0:
                        operations.append(
                            combinators.Apply(e.function, len(e.args))
                        )
                        scope = []
                    else:
                        operations.append(
                            combinators.Parallel(
                                combinators.Apply(e.function, len(e.args)),
                                combinators.Identity(n),
                            )
                        )
                        scope = scope[len(e.args):]
                else:
                    raise NotImplementedError()

                scope = [e] + scope

        return combinators.Serial(*operations)

def differentiable_function(f):
    signature = inspect.signature(f)
    symbolic_args = list(map(expr.Variable, signature.parameters))
    return Function(symbolic_args, f(*symbolic_args)).transform()



class Combinator:
    pass

class Identity(Combinator):
    def __init__(self, n):
        self.n = n

    @property
    def inputs(self):
        return self.n

    @property
    def outputs(self):
        return self.n

    def __call__(self, *scope):
        assert len(scope) == self.n
        def backwards(*gradients):
            assert len(gradients) == self.n
            return tuple(gradients)
        return tuple(scope), backwards

    def __repr__(self):
        return "Identity(%i)" % self.n

class Serial(Combinator):
    def __init__(self, *xs):
        assert len(xs) >= 1
        self.xs = xs

    @property
    def inputs(self):
        return self.xs[0].inputs

    @property
    def outputs(self):
        return self.xs[-1].outputs

    def __call__(self, *scope):
        backs = []
        intermediate = scope
        for fn in self.xs:
            intermediate, back = fn(*intermediate)
            backs.append(back)
        def backward(*gradient):
            intermediate = gradient
            for back in reversed(backs):
                intermediate = back(*intermediate)
            return intermediate
        return intermediate, backward

    def __repr__(self):
        return "Serial(%s)" % (', '.join(map(repr, self.xs)))

class Parallel(Combinator):
    def __init__(self, *xs):
        self.xs = xs

    @property
    def inputs(self):
        return sum([x.inputs for x in self.xs])

    @property
    def outputs(self):
        return sum([x.outputs for x in self.xs])

    def __call__(self, *scope):
        outputs = []
        backs = []
        for x in self.xs:
            output, back = x(*scope[:x.inputs])
            outputs.append(output)
            backs.append(back)
            scope = scope[x.inputs:]
        assert len(scope) == 0
        def backward(*gradients):
            outputs = []
            for back, x in zip(backs, self.xs):
                output = back(*gradients[:x.outputs])
                outputs.append(output)
                gradients = gradients[x.outputs:]
            assert len(scope) == 0
            result = tuple([x for xs in outputs for x in xs])
            return result
        return tuple([x for xs in outputs for x in xs]), backward

    def __repr__(self):
        return "Parallel(%s)" % (', '.join(map(repr, self.xs)))

class Discard(Combinator):
    def __init__(self, *mask):
        self.mask = mask

    @property
    def inputs(self):
        return len(self.mask)

    @property
    def outputs(self):
        return len(filter(lambda x: not x, self.mask))

    def __call__(self, *scope):
        assert len(scope) == len(self.mask)
        output = []
        for m, x in zip(self.mask, scope):
            if not m:
                output.append(x)
        def backward(*gradients):
            raise NotImplementedError()
        return tuple(output), backward

    def __repr__(self):
        return "Discard(%s)" % (', '.join(map(repr, self.mask)))

class Duplicate(Combinator):
    def __init__(self, *mask):
        self.mask = mask

    @property
    def inputs(self):
        return len(self.mask)

    @property
    def outputs(self):
        return len(self.mask) + len(filter(lambda x: x, self.mask))

    def __call__(self, *scope):
        assert len(scope) == len(self.mask)
        output = []
        for m, x in zip(self.mask, scope):
            output.append(x)
            if m:
                output.append(x)
        def backward(*gradients):
            raise NotImplementedError()
        return tuple(output), backward

    def __repr__(self):
        return "Duplicate(%s)" % (', '.join(map(repr, self.mask)))

class Permutation(Combinator):
    def __init__(self, *indices):
        assert list(sorted(indices)) == list(range(len(indices)))
        self.indices = indices
        self.inverse = []
        for i in range(len(self.indices)):
            self.inverse.append(self.indices.index(i))

    @property
    def inputs(self):
        return len(self.indices)

    @property
    def outputs(self):
        return len(self.indices)

    def __call__(self, *scope):
        assert len(scope) == len(self.indices)
        output = []
        for index in self.indices:
            output.append(scope[index])
        def backward(*gradients):
            output = []
            for i in self.inverse:
                output.append(gradients[i])
            return output
        return tuple(output), backward

    def __repr__(self):
        return "Permutation(%s)" % (', '.join(map(str, self.indices)))

class Apply(Combinator):
    def __init__(self, function, n):
        self.function = function
        self.n = n

    @property
    def inputs(self):
        return self.n

    @property
    def outputs(self):
        return 1

    def __call__(self, *scope):
        assert len(scope) == self.n
        output, back = self.function(*scope)
        return (output,), back

    def __repr__(self):
        return "Apply(%s, %d)" % (repr(self.function), self.n)



from cognite import expr
import mxnet as mx

def forward(activations, weights):
    output = mx.ndarray.dot(activations, weights)
    def backwards(gradients):
        activation_gradients = mx.ndarray.dot(gradients, weights.T)
        weight_gradients = mx.ndarray.dot(activations.T, gradients)
        return (activation_gradients, weight_gradients)
    return output, backwards

def linear(x, weights):
    if isinstance(x, expr.Constant) and isinstance(weights, expr.Constant):
        return expr.Constant(forward(x.value, weights.value)[0])
    else:
        return expr.Apply(expr.Function('linear', forward), [x, weights])
